import bisect
import functools
from utils_ak.dict import *
from utils_ak.loguru import *
from loguru import logger

# NOTE: WORKING WITH SINGLE-PARTITIONED KAFKA TOPICS


def get_single_topic_partition(kafka_consumer, topic):
    if len(kafka_consumer.assignment()) == 0:
        # init kafka and reset offsets
        record = next(kafka_consumer)
        partitions = list(kafka_consumer.assignment())

        record_partition = [
            p
            for p in partitions
            if p.topic == record.topic and p.partition == record.partition
        ][0]
        kafka_consumer.seek(record_partition, max(record.offset - 1, 0))

    partitions = list(kafka_consumer.assignment())

    topic_partitions = [p for p in partitions if p.topic == topic]
    assert len(topic_partitions) == 1
    return topic_partitions[0]


def get_record_by_offset(kafka_consumer, topic, offset):
    # fetch first to init
    if offset < 0:
        offset = offset % get_end_offset(kafka_consumer, topic)

    partition = get_single_topic_partition(kafka_consumer, topic)
    kafka_consumer.seek(partition, offset)
    return next(kafka_consumer)


def get_end_offset(kafka_consumer, topic):
    partition = get_single_topic_partition(kafka_consumer, topic)
    return kafka_consumer.end_offsets(kafka_consumer.assignment())[partition]


@functools.total_ordering
class _KafkaRecord:
    def __init__(self, record, lt_key):
        self.record = record
        self.lt_key = lt_key

    def __lt__(self, other):
        return self.lt_key(self.record) < self.lt_key(other.record)


class KafkaConsumerAsList:
    def __init__(self, kafka_consumer, topic, lt_key):
        self.kafka_consumer = kafka_consumer
        self.topic = topic
        self._length = None
        self.lt_key = lt_key

    def __len__(self):
        if not self._length:
            self._length = get_end_offset(self.kafka_consumer, self.topic)
        return self._length

    def __getitem__(self, item):
        if not isinstance(item, int):
            raise KeyError
        item = item % len(self)
        record = get_record_by_offset(self.kafka_consumer, self.topic, item)
        return self.cast_list_record(record)

    def cast_list_record(self, record):
        if isinstance(record, dict):
            record = dotdict(record)
        return _KafkaRecord(record, self.lt_key)


TOPIC = "datasets__60705b5edaf0f2d1693a39c6"


def test():
    from kafka import KafkaConsumer

    consumer = KafkaConsumer(
        TOPIC,
        bootstrap_servers="localhost:9092",
        auto_offset_reset="earliest",
        enable_auto_commit=True,
    )
    consumer.assignment()
    print(get_record_by_offset(consumer, TOPIC, 0))
    print(get_record_by_offset(consumer, TOPIC, 1))
    print(get_record_by_offset(consumer, TOPIC, 0))
    print(get_record_by_offset(consumer, TOPIC, -1))


def test_kafka_bisect_left():
    from kafka import KafkaConsumer

    consumer = KafkaConsumer(
        TOPIC,
        bootstrap_servers="localhost:9092",
        auto_offset_reset="earliest",
        enable_auto_commit=True,
    )

    consumer_as_list = KafkaConsumerAsList(
        consumer, TOPIC, lt_key=lambda record: record.timestamp
    )

    print(
        bisect.bisect_left(
            consumer_as_list,
            consumer_as_list.cast_list_record({"timestamp": 1617976164019}),
        )
    )


if __name__ == "__main__":
    test()
    test_kafka_bisect_left()



from utils_ak.coder.coders import *
from utils_ak.coder.coders.tests.coder_tests import _test_coder


def test_coders():
    _test_coder(JsonCoder(), assert_equal=False)
    _test_coder(MsgPackCoder(), assert_equal=False)


if __name__ == "__main__":
    test_coders()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['utils_ak',
 'utils_ak.architecture',
 'utils_ak.architecture.finite_state_machine',
 'utils_ak.architecture.func',
 'utils_ak.architecture.integrity_guard',
 'utils_ak.architecture.operation',
 'utils_ak.architecture.sink',
 'utils_ak.async_tools',
 'utils_ak.block_tree',
 'utils_ak.block_tree.pushers',
 'utils_ak.builtin',
 'utils_ak.callback_timer',
 'utils_ak.clickhouse',
 'utils_ak.clickhouse.config',
 'utils_ak.clock',
 'utils_ak.coder',
 'utils_ak.coder.coders',
 'utils_ak.coder.coders.json',
 'utils_ak.coder.coders.msgpack',
 'utils_ak.coder.coders.tests',
 'utils_ak.color',
 'utils_ak.config',
 'utils_ak.dag',
 'utils_ak.deployment',
 'utils_ak.deployment.controller',
 'utils_ak.deployment.controller.controllers',
 'utils_ak.deployment.controller.controllers.inbox',
 'utils_ak.deployment.controller.controllers.tests',
 'utils_ak.deployment.examples',
 'utils_ak.deployment.examples.hello_world',
 'utils_ak.deployment.tests',
 'utils_ak.dict',
 'utils_ak.exceptions',
 'utils_ak.fluid_flow',
 'utils_ak.fluid_flow.actors',
 'utils_ak.fluid_flow.tests',
 'utils_ak.git',
 'utils_ak.hash',
 'utils_ak.id_generator',
 'utils_ak.id_generator.id_generators',
 'utils_ak.iteration',
 'utils_ak.job_orchestrator',
 'utils_ak.job_orchestrator.scripts',
 'utils_ak.job_orchestrator.tests',
 'utils_ak.job_orchestrator.tests.config',
 'utils_ak.job_orchestrator.worker',
 'utils_ak.job_orchestrator.worker.sample_worker',
 'utils_ak.job_orchestrator.worker.tests',
 'utils_ak.jsonrpc.clients',
 'utils_ak.jsonrpc.clients.archive',
 'utils_ak.jupyter',
 'utils_ak.kafka',
 'utils_ak.kafka.archive',
 'utils_ak.loguru',
 'utils_ak.message_queue',
 'utils_ak.message_queue.brokers',
 'utils_ak.message_queue.brokers.brokers',
 'utils_ak.message_queue.clients',
 'utils_ak.mongoengine',
 'utils_ak.naming',
 'utils_ak.networking',
 'utils_ak.networking.proxy',
 'utils_ak.nn',
 'utils_ak.numeric',
 'utils_ak.openpyxl',
 'utils_ak.os',
 'utils_ak.pandas',
 'utils_ak.plot',
 'utils_ak.portion',
 'utils_ak.properties',
 'utils_ak.properties.archive',
 'utils_ak.proxy',
 'utils_ak.re',
 'utils_ak.reflexion',
 'utils_ak.screen',
 'utils_ak.screen.archive',
 'utils_ak.screen.tests',
 'utils_ak.simple_event_manager',
 'utils_ak.simple_microservice',
 'utils_ak.simple_microservice.examples',
 'utils_ak.simple_vector',
 'utils_ak.split_file',
 'utils_ak.state',
 'utils_ak.state.provider',
 'utils_ak.state.provider.providers',
 'utils_ak.str',
 'utils_ak.streaming',
 'utils_ak.streaming.windowing',
 'utils_ak.system',
 'utils_ak.time',
 'utils_ak.tqdm',
 'utils_ak.zmq',
 'utils_ak.zmq.benchmark']

package_data = \
{'': ['*'],
 'utils_ak.config': ['sample_configs/*'],
 'utils_ak.deployment.examples.hello_world': ['inbox/*'],
 'utils_ak.job_orchestrator': ['data/docker-compose/603384f556f388e8aeceb221/*',
                               'data/docker-compose/603385670c7c448055c8a574/*',
                               'data/docker-compose/60338b8dcfb2b0d1518ab942/*',
                               'data/docker-compose/60338c65a934f591503fa342/*',
                               'data/docker-compose/60338c7b3ceb6807bc8bf137/*',
                               'data/docker-compose/60338cd3d10015ef72308dca/*',
                               'data/docker-compose/6033ac8e50a3b29e09cf4b47/*',
                               'data/docker-compose/6033acb73993552852600347/*',
                               'data/docker-compose/6033acdd0ec1968d521e6ff0/*',
                               'data/docker-compose/6033ad356ad37d72409f823c/*',
                               'data/docker-compose/6033ad6227fcab41c088641c/*',
                               'data/docker-compose/6033ad832ef3c2e3d6f707e4/*'],
 'utils_ak.job_orchestrator.tests.config': ['configs/*'],
 'utils_ak.screen': ['logs/*', 'screen_logs/*']}

setup_kwargs = {
    'name': 'utils-ak',
    'version': '0.1.13',
    'description': 'Lots of small utils',
    'long_description': None,
    'author': 'akadaner',
    'author_email': 'arseniikadaner@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



# builtin in python 3.9
# todo: migrate to python 3.9
def trim(s, beg=None, end=None):
    if beg is not None and s.startswith(beg):
        s = s[len(beg) :]
    if end is not None and s.endswith(end):
        s = s[: -len(end)]
    return s



import time
import asyncio
from utils_ak.coder import JsonCoder
from utils_ak.id_generator import IncrementalIDGenerator


class WebSocketsClient:
    def __init__(self, websocket_client, id_generator=None):
        self.websocket_client = websocket_client
        self.responses = {}  # {<id>: response}
        self.coder = JsonCoder(encoding=None)

        self.active = True
        self.id_generator = id_generator or IncrementalIDGenerator()

    def _prepare_request(self, method, generate_id=True, **kwargs):
        request = {"jsonrpc": "2.0", "params": kwargs, "method": method}
        if generate_id:
            request["id"] = self.id_generator.gen_id()
        return request

    async def execute(self, method, notify=False, **kwargs):
        """
        :param request: {"jsonrpc": "2.0", "method": "post.like", "params": {"post": "12345"}, "id": 1}
        """
        request = self._prepare_request(method, generate_id=not notify, **kwargs)
        await self.websocket_client.send(self.coder.encode(request))

        if not notify:
            while True:
                if request["id"] in self.responses:
                    return self.responses.pop(request["id"])
                await asyncio.sleep(0.001)

    async def start_receiving_loop(self):
        while True:
            if not self.active:
                return

            try:
                response_text = await asyncio.wait_for(
                    self.websocket_client.recv(), timeout=0.1
                )
            except asyncio.TimeoutError:
                await asyncio.sleep(0)
                continue
            data = self.coder.decode(response_text)
            self.responses[data["id"]] = data
            await asyncio.sleep(0)

    def stop_receiving_loop(self):
        self.active = False



from utils_ak.block_tree.pushers import *
from utils_ak.block_tree.parallelepiped_block import ParallelepipedBlock


class BlockMaker:
    def __init__(
        self, root_obj="root", default_push_func=stack_push, block_factory=None, **props
    ):
        self.block_factory = block_factory or ParallelepipedBlock

        if isinstance(root_obj, str):
            self.root = self.block_factory(root_obj, **props)
        elif isinstance(root_obj, ParallelepipedBlock):
            assert len(props) == 0  # not supported case
            self.root = root_obj
        else:
            raise Exception("Unknown root type")

        self.blocks = [self.root]
        self.default_push_func = default_push_func

    def create_block(self, block_obj, **kwargs):
        return self.block_factory(block_obj, **kwargs)

    def copy(self, block, with_children=True, with_props=False, prop_keys=None):
        res = self.create_block(block.props["cls"], **block.props.relative_props)

        if with_children:
            for child in block.children:
                res.add_child(self.copy(child))

        if with_props:
            props = block.props.all()
            prop_keys = prop_keys or []
            if prop_keys:
                props = {k: v for k, v in props if k in prop_keys}
            props = {k: v for k, v in props.items() if v is not None}
            res.props.update(**props)
        return res

    def make(self, block_obj=None, push_func=None, push_kwargs=None, **kwargs):
        push_func = push_func or self.default_push_func
        push_kwargs = push_kwargs or {}

        if isinstance(block_obj, str) or block_obj is None:
            block = self.create_block(block_obj, **kwargs)
        elif isinstance(block_obj, ParallelepipedBlock):
            block = block_obj
            block.props.update(**kwargs)
        else:
            raise Exception("Unknown block obj type")

        push_func(self.blocks[-1], block, **push_kwargs)
        return BlockMakerContext(self, block)


class BlockMakerContext:
    def __init__(self, maker, block):
        self.maker = maker
        self.block = block

    def __enter__(self):
        self.maker.blocks.append(self.block)
        return self.block

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.maker.blocks.pop()


def init_block_maker(root_obj, default_push_func=stack_push, **kwargs):
    block_maker = BlockMaker(root_obj, default_push_func, **kwargs)
    return block_maker, block_maker.make


def test_block_maker1():
    maker, make = init_block_maker("root", axis=0)
    make("a", size=[1, 0])
    make("b", size=[5, 0])
    make(maker.create_block("c", size=[2, 0]), test=5)
    print(maker.root)
    print(maker.root["c"].props.all())


def test_block_maker2():
    maker, make = init_block_maker("root", axis=1)
    with make("a1", size=[0, 3]):
        with make("b1", size=[5, 0]):
            make("c1", size=[2, 0])
    with make("a2", size=[0, 2]):
        make("b2")

    print(maker.root)


def test_copy():
    maker, make = init_block_maker("root")
    with make("a1", x=[1, 1], size=[1, 1], push_func=add_push):
        make("b1", size=[1, 1])
        with make("b2", size=[1, 1]):
            make("c1", size=[1, 1])
    print(maker.root)
    print(maker.copy(maker.root["a1"]["b2"]))
    print(maker.copy(maker.root["a1"]["b2"], with_props=True))


if __name__ == "__main__":
    test_block_maker1()
    test_block_maker2()
    test_copy()



import time
import multiprocessing
from loguru import logger
from mongoengine import connect as connect_to_mongodb

from utils_ak.simple_microservice import run_listener_async
from utils_ak.deployment import *
from utils_ak.loguru import configure_loguru_stdout

from utils_ak.job_orchestrator.models import *
from utils_ak.job_orchestrator.job_orchestrator import JobOrchestrator
from utils_ak.job_orchestrator.tests.test_monitor import run_monitor
from utils_ak.job_orchestrator.tests.config.config import config
from utils_ak.job_orchestrator.worker.sample_worker.main import MAIN


def create_new_job(payload, python_main=MAIN, drop_collections=True):
    configure_loguru_stdout("DEBUG")
    connect_to_mongodb(host=config.MONGODB_HOST, db=config.MONGODB_DB)
    logger.info("Connected to mongodb")
    time.sleep(2)

    if drop_collections:
        logger.info("Dropping Job and Worker collections...")
        Job.drop_collection()
        Worker.drop_collection()

    logger.debug("Creating new job...")
    payload = dict(payload)
    payload.update(
        {
            "message_broker": config.TRANSPORT,
        }
    )

    job = Job(
        type="test",
        payload=payload,
        runnable={
            "image": "akadaner/test-worker",
            "python_main": python_main,
        },
        running_timeout=60,
    )
    job.save()


def run_job_orchestrator(payload=None):
    configure_loguru_stdout("DEBUG")
    connect_to_mongodb(host=config.MONGODB_HOST, db=config.MONGODB_DB)
    logger.info("Connected to mongodb")
    controller = ProcessController()

    run_listener_async("job_orchestrator", message_broker=config.TRANSPORT)
    logger.info('Running job orchestrator...')
    job_orchestrator = JobOrchestrator(controller, config.TRANSPORT)
    multiprocessing.Process(target=run_monitor).start()
    if payload:
        multiprocessing.Process(target=create_new_job, args=(payload,)).start()
    job_orchestrator.run()


def run_success():
    run_job_orchestrator({"type": "batch"})


def run_stalled():
    run_job_orchestrator({"type": "batch", "stalled_timeout": 600})


def run_timeout():
    run_job_orchestrator({"type": "batch", "running_timeout": 20})


def run_failure():
    run_job_orchestrator({"type": "batch", "initializing_timeout": 600})


def run():
    run_job_orchestrator()


if __name__ == "__main__":
    run_success()
    # run_stalled()
    # run_timeout()
    # run_failure()
    # run()



import msgpack
import decimal
import datetime as datetime_module

from utils_ak.coder import Coder

from utils_ak.coder.binary import (
    pack_datetime,
    pack_decimal,
    pack_time,
    unpack_datetime,
    unpack_decimal,
    unpack_time,
    pack_date,
    unpack_date,
)

EXT_DATE = 1
EXT_DATETIME = 2
EXT_DECIMAL = 3
EXT_TIME = 4


def default(obj):
    if isinstance(obj, datetime_module.datetime):
        return msgpack.ExtType(EXT_DATETIME, pack_datetime(obj))
    elif isinstance(obj, datetime_module.date):
        return msgpack.ExtType(EXT_DATE, pack_date(obj))
    elif isinstance(obj, datetime_module.time):
        return msgpack.ExtType(EXT_TIME, pack_time(obj))
    elif isinstance(obj, decimal.Decimal):
        return msgpack.ExtType(EXT_DECIMAL, pack_decimal(obj))
    else:
        # Wuh-woh
        raise TypeError(
            u"Cannot encode value of type {} to MessagePack: {}".format(
                type(obj).__name__, obj
            )
        )


def ext_hook(code, data):
    """
    Decodes our custom extension types
    """
    if code == EXT_DATETIME:
        return unpack_datetime(data)
    elif code == EXT_DATE:
        return unpack_date(data)
    elif code == EXT_TIME:
        return unpack_time(data)
    elif code == EXT_DECIMAL:
        return unpack_decimal(data)
    else:
        raise TypeError(
            u"Cannot decode unknown extension type {} from MessagePack".format(code)
        )


class MsgPackCoder(Coder):
    def encode(self, obj):
        return msgpack.dumps(obj, default=default, use_bin_type=True)

    def decode(self, obj):
        return msgpack.loads(obj, ext_hook=ext_hook, raw=False)


if __name__ == "__main__":
    from utils_ak.coder import _test_coder

    test_coder(MsgPackCoder())



from utils_ak.fluid_flow import *


def test_pipe_switch_1():
    from utils_ak.fluid_flow.actors.container import Container

    ci1 = Container("I1")
    ci2 = Container("I2")
    co1 = Container("O1")
    co2 = Container("O2")

    pipe_connect(ci1, co1, "1")
    pipe_connect(ci2, co2, "2")

    for node in ci1.iterate():
        print(node)
    for node in ci2.iterate():
        print(node)
    print()

    pipe_switch(co1, co2, "in")

    for node in ci1.iterate():
        print(node)
    for node in ci2.iterate():
        print(node)
    print()

    pipe_switch(co1, co2, "in")
    for node in ci1.iterate():
        print(node)
    for node in ci2.iterate():
        print(node)
    print()

    pipe_switch(ci1, ci2, "out")

    for node in ci1.iterate():
        print(node)
    for node in ci2.iterate():
        print(node)
    print()


def test_pipe_switch_2():
    from utils_ak.fluid_flow.actors.container import Container

    ci1 = Container("I1")
    co1 = Container("O1")
    co2 = Container("O2")

    pipe_connect(ci1, co1)

    for node in ci1.iterate():
        print(node)
    print()
    pipe_switch(co1, co2, "in")

    for node in ci1.iterate():
        print(node)
    print()


if __name__ == "__main__":
    test_pipe_switch_1()
    test_pipe_switch_2()



import logging

from utils_ak.zmq import endpoint
from utils_ak.simple_microservice import SimpleMicroservice, run_listener_async
from utils_ak.loguru import configure_loguru_stdout


class Publisher(SimpleMicroservice):
    def __init__(self, *args, **kwargs):
        super().__init__("Publisher", *args, **kwargs)
        self.add_timer(self.timer_function, 2)

    def timer_function(self):
        self.publish("collection", "")


if __name__ == "__main__":
    configure_loguru_stdout()
    run_listener_async(
        "collection",
        message_broker=(
            "zmq",
            {
                "endpoints": {
                    "collection": {
                        "type": "sub",
                        "endpoint": endpoint("localhost", 6554),
                    }
                }
            },
        ),
    )
    Publisher(
        message_broker=(
            "zmq",
            {
                "endpoints": {
                    "collection": {
                        "type": "sub",
                        "endpoint": endpoint("localhost", 6554),
                    }
                }
            },
        )
    ).run()



from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.actors.pipe import PipeMixin, Pipe
from utils_ak.fluid_flow.calculations import *


class Hub(Actor):
    def __init__(self, name):
        super().__init__(name)

    def update_pressure(self, ts):
        # todo: hardcode. Need updated children pressures. Better solution? Update pressure top-down?
        for pipe in self.children:
            pipe.child.update_pressure(ts)

        parent_items = [pipe.current_item for pipe in self.parents]
        assert len(set(parent_items)) == 1
        item = parent_items[0]
        children_with_item = [
            pipe for pipe in self.children if pipe.current_item == item
        ]

        if any(pipe.pressures["in"] is None for pipe in children_with_item):
            for pipe in self.parents:
                pipe.pressures["in"] = None
        else:
            total_output_pressure = sum(
                pipe.pressures["in"] for pipe in children_with_item
            )
            left = total_output_pressure

            for pipe in self.parents:
                pipe.pressures["in"] = nanmin([left, pipe.pressures["out"]])
                left -= nanmin([left, pipe.pressures["out"]])

    def update_speed(self, ts):
        assert all(isinstance(pipe, Pipe) for pipe in self.parents + self.children)

        parent_items = [pipe.current_item for pipe in self.parents]
        assert len(set(parent_items)) == 1
        item = parent_items[0]
        children_with_item = [
            pipe for pipe in self.children if pipe.current_item == item
        ]

        total_input_speed = sum(pipe.current_speed for pipe in self.parents)
        left = total_input_speed

        for pipe in self.children:
            if pipe in children_with_item:
                pipe.pressures["out"] = nanmin([left, pipe.pressures["in"]])
                left -= nanmin([left, pipe.pressures["in"]])
            else:
                pipe.pressures["out"] = 0

    def __str__(self):
        return f"Hub: {self.name}"
